//
// Created by philipp on 30.11.17.
//

#include <boost/asio/ip/tcp.hpp>
#include <iostream>
#include "../include/PlatoonProtocolLib/FollowerVehicle.h"

namespace platoonProtocol
{

using namespace std::chrono_literals;

constexpr networking::time::Duration FollowerVehicle::HEARTBEAT_INTERVAL;
constexpr networking::time::Duration FollowerVehicle::BROADCAST_TIMEOUT;
constexpr networking::time::Duration FollowerVehicle::RESPONSE_TIMEOUT;

FollowerVehicle::Ptr FollowerVehicle::create(networking::Networking & net, const NetworkInfo & info)
{
    return std::make_shared<FollowerVehicle>(PrivateTag{}, net, info);
}

FollowerVehicle::FollowerVehicle(PrivateTag, networking::Networking & net, const NetworkInfo & info)
    : Vehicle(net, info, Role::FOLLOWER)
{
    heartbeatTimer = networking::time::Timer::create(net);
}

void FollowerVehicle::stop()
{
    Vehicle::stop();

    stopHeartbeat();

    stopBroadcastTimer();
}

PlatoonConfig FollowerVehicle::getPlatoonConfig() const noexcept
{
    return platoonConfig.load();
}

void FollowerVehicle::doCreatePlatoon()
{
    if (state != State::IDLE)
        return;

    Vehicle::doCreatePlatoon();

    // We cycle through every available endpoint and try to make a create-platoon-request.
    // Note that we'll be actually starting with the first element in the vector
    // by setting the iterator to end() due to utils::cycle().
    vehiclesToRequestIter = vehiclesToRequest.end();
    sendPlatoonCreateRequestToNextVehicle();
}

void FollowerVehicle::doRunPlatoon()
{
    if (state != State::CREATING_PLATOON)
        return;

    Vehicle::doRunPlatoon();

    startHeartbeat();

    startBroadcastTimer();
}

void FollowerVehicle::doLeavePlatoon()
{
    if (state == State::IDLE)
        return;

    Vehicle::doLeavePlatoon();

    sendLeavePlatoonMessage();
}

void FollowerVehicle::receiveMessage(const PlatoonMessage & message)
{
    switch (message.messageType)
    {
        case messageTypes::LV_BROADCAST:
            receiveBroadcast(message);
            break;

        case messageTypes::FV_REQUEST:
            receivePlatoonCreateRequest(message);
            break;

        case messageTypes::LV_ACCEPT:
        case messageTypes::REJECT:
            receiveResponse(message);
            break;

        default:
            // ignore rest
            break;
    }
}

void FollowerVehicle::sendPlatoonCreateRequestToNextVehicle()
{
    if (state != State::CREATING_PLATOON || vehiclesToRequest.empty())
        return;

    utils::cycle(vehiclesToRequestIter, vehiclesToRequest);

    auto vehicleToRequest = *vehiclesToRequestIter;
    sendPlatoonCreateRequest(vehicleToRequest);

    auto self = shared_from_this();
    // Create a new one to make sure we're never getting a busy exception.
    responseTimer = networking::time::Timer::create(net);
    responseTimer->startTimeout(
        RESPONSE_TIMEOUT,
        [self]
        {
            log("Response Timeout!\n");
            self->sendPlatoonCreateRequestToNextVehicle();
        });
}

void FollowerVehicle::sendPlatoonCreateRequest(VehicleId vehicleToRequest)
{
    // We'll send a request indicating that we want to be the follower.
    sendMessage(PlatoonMessage::followerRequest(myInfo.vehicleId, vehicleToRequest), 3s);
}

void FollowerVehicle::receivePlatoonCreateRequest(const PlatoonMessage & request)
{
    if (request.dstVehicle != myInfo.vehicleId)
        return;

    // Always reject because only Leaders can take requests.
    sendMessage(PlatoonMessage::rejectResponse(myInfo.vehicleId, request.srcVehicle), 3s);
}

void FollowerVehicle::startHeartbeat()
{
    auto self = shared_from_this();
    heartbeatTimer->startPeriodicTimeout(
        HEARTBEAT_INTERVAL,
        [self]
        {
            self->sendMessage(PlatoonMessage::heartbeatMessage(self->myInfo.vehicleId, self->leader, self->platoonId),
                              HEARTBEAT_INTERVAL);
        });
}

void FollowerVehicle::stopHeartbeat()
{
    heartbeatTimer->stop();
}

void FollowerVehicle::startBroadcastTimer()
{
    auto self = shared_from_this();
    broadcastTimer = networking::time::Timer::create(net);
    broadcastTimer->startTimeout(
        BROADCAST_TIMEOUT,
        [self]
        {
            log("Broadcast Timeout!\n");
            self->doLeavePlatoon();
        });
}

void FollowerVehicle::stopBroadcastTimer()
{
    broadcastTimer->stop();
}

void FollowerVehicle::receiveBroadcast(const PlatoonMessage & broadcast)
{
    if (state != State::RUNNING_PLATOON)
        return;

    if (broadcast.srcVehicle != leader ||
        broadcast.platoonId != platoonId)
        return;

    stopBroadcastTimer();
    startBroadcastTimer();

    PlatoonConfig newConfig{broadcast.platoonSpeed, broadcast.innerPlatoonDistance};
    if (platoonConfig.load() != newConfig)
    {
        platoonConfig = newConfig;
        onPlatoonConfigUpdatedCallback();
    }
}

void FollowerVehicle::receiveResponse(const PlatoonMessage & response)
{
    if (state != State::CREATING_PLATOON ||
        response.dstVehicle != myInfo.vehicleId ||
        response.srcVehicle != *vehiclesToRequestIter)
        return;

    responseTimer->stop();

    if (response.messageType == messageTypes::LV_ACCEPT)
    {
        leader = response.srcVehicle;
        platoonId = response.platoonId;
        doRunPlatoon();
    }
    else
    {
        sendPlatoonCreateRequestToNextVehicle();
    }
}

void FollowerVehicle::sendLeavePlatoonMessage()
{
    sendMessage(PlatoonMessage::leaveMessage(myInfo.vehicleId, leader, platoonId), HEARTBEAT_INTERVAL);
}

void FollowerVehicle::addVehicleToRequest(VehicleId vehicleId)
{
    if (vehicleId == 0)
        throw std::invalid_argument{"Invalid vehicle ID '0'"};

    auto self = shared_from_this();
    net.callLater(
        [self, vehicleId]
        { self->vehiclesToRequest.push_back(vehicleId); });
}

FollowerVehicle::Ptr FollowerVehicle::shared_from_this()
{
    return std::static_pointer_cast<FollowerVehicle>(Vehicle::shared_from_this());
}

}